 /**********************
 *                     *
 *   Geometry object   *
 *    Constructors     *
 *                     *
 **********************/

var defaultEpsilon = 1e-5;

function Coordinates (fX, fY, fZ) {

  this.length = 9;
  this.x = fX;                            //float (r/w)
  this.y = fY;                            //float (r/w)
  this.z = fZ;                            //float (r/w)
  this.epsilon = defaultEpsilon;          //float (r/w)
  this.compare = coordCompare;            //method
  this.dist = coordDist;                  //method
  this.toString = coordToString;          //method
  this.toArray = coordToArray;            //method
  this.setEpsilon = coordSetEpsilon;      //method
}

function coordCompare (a, b, c, d) {

  var fEpsilon, bx, by, bz;

  if (a.constructor == Coordinates){
    fEpsilon = ((b != undefined)? Math.pow(10, -b): this.epsilon);
    with (Math) {
      bx = abs(this.x - a.x) <= fEpsilon;
      by = abs(this.y - a.y) <= fEpsilon;
      bz = (this.z != undefined || a.z != undefined)? abs(this.z - a.z) <= epsilon: true;
    }
  }
  else { 
    fEpsilon = ((d != undefined)? Math.pow(10, -d): this.epsilon);
    with (Math) {
      bx = abs(this.x - a) <= fEpsilon;
      by = abs(this.y - b) <= fEpsilon;
      bz = (this.z != undefined)? abs(this.z - c) <= fEpsilon: true;
    }
  }
  return((bx && by && bz)? true: false);
}

function coordDist (a, b, c){

  var dx,dy,dz,dist,bTriD;

  bTriD = (this.z != undefined) && (c != undefined);
  if (a.constructor==Coordinates){
    dx = this.x - a.x;
    dy = this.y - a.y;
    dz = bTriD? this.z - a.z: 0;
  }
  else { 
    dx = this.x - a;
    dy = this.y - b;
    dz = bTriD? this.z - a.z: 0;
  }
  with(Math) dist = sqrt(pow(dx,2)+pow(dy,2)+pow(dz,2));
  return(dist);
}

function coordToString(sBracket){

  var s = this.x + "," + this.y;
  s += (this.z != undefined) ? "," + this.z : "";
  s = (sBracket != undefined) ? sBracket.substr(0,1) + s + sBracket.substr(1,1): s;
  return(s);
}

//sBracket constants:
var jsRound = "()";
var jsSquare = "[]";
var jsBrace = "{}";
var jsAngle = "<>";
var jsPipe = "||";
var jsNone = "";
//can also use custom string, eg "[)"

function coordToArray(){

  var s = new Array(this.x, this.y, this.z);
  return(s);
}

function coordSetEpsilon (a, b) {
  this.epsilon = Math.pow((b != undefined? b: 10), -a);
}

var Origin = new Coordinates(0,0,0);

function Line2D(fm, fc){

  this.length = 6;
  this.m = fm;
  this.intercept = fc;
  this.epsilon = defaultEpsilon;
  this.calcY = lineCalcY;
  this.calcX = lineCalcX;
  this.angleOfIntercept = lineaoi;
  this.perpDist = linePD;
  this.isOn = lineIsOn;
  this.toString = lineToString
}

function lineCalcY(x){
  return(x * this.m + this.intercept);
}

function lineCalcX(y){
  return((y - this.intercept) / this.m);
}

function lineaoi(a){
  if (a.constructor == Line2D) return(Math.atan(this.m) + Math.atan(a.m));
  else return(NaN);
}

function linePD(a,b){

  var perpC, px, py, dist, perpM = -1 / this.m;
  if (a.constructor == Coordinates) {
    perpC = a.y - perpM * a.x;
    px = (perpC - this.intercept) / (this.m - perpM);
    py = this.calcY(px);
    dist = a.dist(px,py);
  }
  else {
    perpC = b - perpM * a;
    px = (perpC - this.intercept) / (this.m - perpM);
    py = this.calcY(px);
    with (Math) dist = sqrt(pow(a - px, 2) + pow(b - py, 2))
  }
  return(dist);
}

function lineIsOn(a,b,c){
  
  var fEpsilon,bMatch;

  if (a.constructor == Coordinates) {
    with (Math) {
      fEpsilon = (b != undefined)? pow(10,-b): min(a.epsilon, this.epsilon);
      fMatch = abs(this.calculateY(a.x)-a.y);
    }
  }
  else {
    with (Math) {
      fEpsilon = (c != undefined)? pow(10,-c): this.epsilon;
      fMatch = abs(this.calculateY(a)-b);
    }
  }
  return((fMatch <= (fEpsilon * 2))? true: false);
}

function lineToString() {
  
  var sRet = "y = ";
  switch (this.m) {
  case -1:
    sRet += "-x";
    break;
  case 0:
    sRet += this.intercept;
    break;
  case 1:
    sRet += "x";
    break;
  default:
    sRet += this.m + "x";
    break;
  }

  if (this.intercept == 0 || this.m == 0) sRet += "";
  else if (this.intercept > 0) sRet += " + " + this.intercept;
  else sRet += " - " + -this.intercept;
  return(sRet);
}